"""-----------------------------------------------
Cours : TNSI
Titre : L'art du camouflage informatique
Codage : string, boucles for et while, test
Source : GNU/Linux Magazine n°182
Module à installer : Pillow
Objectif : cacher un texte dans une image avec
l'algorithme LSB
Ressources 
- PIL : https://info.blaisepascal.fr/pillow
- A quoi sert yield ? : https://bit.ly/3fI0u9a
- WebGE Wikis : https://webge.synology.me/dokuwiki/doku.php?id=python:accueilpython
Date : 5/2021
Fichier : steganographie.py
-----------------------------------------------"""

#%%
from PIL import Image


def playHead(max, numeroPixel=0, composanteCouleur=0):
    '''
    Tête de lecture sur l'image
    Retourne un tuple composé du numéro de pixel et du 
    numéro associé à la composante couleur (R, V, B) traitée
    Rouge:0 Vert:1 Bleu:2
    :param max: nombre de pixels de l'image
    :param numeroPixel: numéro du pixel dans l'image
    :param composanteCouleur: R:0, V:1, B:2
    '''
    octetCourant = (numeroPixel, composanteCouleur) # tuple renvoyé par yield
    i = 0
    # i et octetCourant sont mémorisés entre deux appels !
    while i < max: # on test si on est à la fin de l'image
        yield octetCourant # on retourne le tuple 'octetCourant' et on sort
        if octetCourant[1] == 2: # on reprendra ici au prochain appel
            octetCourant = (octetCourant[0]+1, 0)
        else:
            octetCourant = (octetCourant[0], octetCourant[1]+1)
        i += 1


def linear2matrix(n, cols):
    '''
    Fournit les numéros de colonne et de ligne d'un pixel
    dans la position linéaire.
    :param n: position linéaire
    :param cols: nombre de colonnes de la matrice
    :return: n° de colonne, n° de ligne
    '''
    return(n % cols, n//cols)


def writeBitInImage(bit, img, head, cols):
    '''
    Ecrit les informations sur le bit de poids faible
    de la composante (R,V ou B) issue de la tête de lecture
    :param bit: bit à écrire
    :param img: image à traiter
    :param head: tête de lecture
    :param cols: largeur de l'image (nb de colonnes de la matrice de pixels)
    :return: la composante modifiée
    '''
    (n_pixel, component) = next(head) # voir testyield.py
    (col, row) = linear2matrix(n_pixel, cols)
    pixel = img.getpixel((col, row))
    new_pixel = (int(bin(pixel[0])[:7]+bit, 2) if component == 0 else pixel[0], int(bin(pixel[1])[:7]+bit, 2)
                 if component == 1 else pixel[1], int(bin(pixel[2])[:7]+bit, 2) if component == 2 else pixel[2])
    img.putpixel((col, row), new_pixel)
    return component


def readBitInImage(img, head, cols):
    '''
    Lit le bit de poids faible de la composante du pixel
    indiqué par la tête de lecture
    :param img: image à traiter
    :param head: tête de lecture
    :param cols: largeur de l'image (nb de colonnes de la matrice de pixels)
    '''
    (n_pixel, component) = next(head)
    (col, row) = linear2matrix(n_pixel, cols)
    pixel = img.getpixel((col, row))
    return (bin(pixel[component])[-1], component)


def lsb(msg, img, filename='img\\out.png', verb=True):
    '''
    Incorpore le message dans l'image et
    sauvegarde le résultat
    :param msg: message à incorporer
    :param img: image à traiter
    :param filename: image traitée sauvegardée dans out.png
    :param verb: ?
    '''
    size = len(msg)
    image = Image.open(img)
    (width, height) = image.size
    nb_pixels = width*height
    head = playHead(nb_pixels)
    encoded = image.copy()

    if nb_pixels < (3*size+6):
        print("Image trop petite pour y intégrer le message")
        exit()

    # Insertion de la taille du message
    b_size = '{:016d}'.format(int(bin(size)[2:]))
    if verb:
        print("Ecriture de la taille : {} soit {} ".format(size, b_size))
    for bit in b_size:
        last_component = writeBitInImage(bit, encoded, head, width)

    # On avance la tête jusqu'au prochain pixel
    while last_component != 2:
        (pixel, last_component) = next(head)

    # Insertion du message
    for c in msg:
        car = '{:08d}'.format(int(bin(ord(c))[2:]))
        if verb:
            print("Ecriture du caractère : {} soit {}b".format(c, car))
        for bit in car:
            last_component = writeBitInImage(bit, encoded, head, width)

        # On avance la tête jusqu'au prochain pixel
        while last_component != 2:
            (pixel, last_component) = next(head)

    encoded.save(filename)


def detect_lsb(img, verb=True):
    '''
    Lit un message dissimulé dans une image
    :param img: image contenant le message
    :param verb: ?
    :return: le message dissimulé dans l'image
    '''
    image = Image.open(img)
    (width, height) = image.size
    nb_pixels = width*height
    head = playHead(nb_pixels)
    size = ''
    msg = ''

    # On récupère la taille du message
    for i in range(16):
        bit, last_component = readBitInImage(image, head, width)
        size += bit
    final_size = int(size, 2)
    if verb:
        print("Taille : {}b soit {} caractères".format(size, final_size))

     # On avance la tête jusqu'au prochain pixel
    while last_component != 2:
        (pixel, last_component) = next(head)

    # On récupère le message
    for i in range(final_size):
        car = ""
        for j in range(8):
            bit, last_component = readBitInImage(image, head, width)
            car += bit
        final_car = chr(int(car, 2))
        msg += final_car
        if verb:
            print("Caractère : {}b soit {}".format(car, final_car))

        # On avance la tête jusqu'au prochain pixel
        while last_component != 2:
            (pixel, last_component) = next(head)

    return msg


if __name__ == "__main__":
    message = "GNU/Linux Magazine"
    lsb(message, "img\\lena.png")
    msg = detect_lsb("img\\out.png")
    print(msg)

# %%
